/* Editor engine */

#ifndef _Ib
#define _Ib 1

#include "misc.h"
#include "rc.h"
#include "vfile.h"

typedef struct buffer B;
typedef struct point P;
typedef struct header H;

struct header
{
	LINK(H) link;
	int seg;
	int hole;
	int ehole;
	int nlines;
};

struct point
{
	LINK(P) link;

	B *b;
	int ofst;
	char *ptr;
	H *hdr;

	int byte;
	int line;
	int col;
	int xcol;
	int valcol;
	int end;

	P **owner;
};

struct buffer
{
	LINK(B) link;
	P *bof;
	P *eof;
	char *name;
	int orphan;
	int count;
	int changed;
	int backup;
	void *undo;
	P *marks[10];             /* Bookmarks */
	OPTIONS o;                /* Options */
	P *oldcur;                /* Last cursor position before orphaning */
	P *oldtop;                /* Last top screen position before orphaning */
	int rdonly;               /* Set for read-only */
	int internal;             /* Set for internal buffers */
	int er;                   /* Error code when file was loaded */
};

extern int force;             /* Set to have final '\n' added to file */
extern int tabwidth;          /* Default tab width */
extern VFILE *vmem;           /* Virtual memory file used for buffer system */
extern char *msgs[];
extern int error;

#define piscol(p) ((p)->valcol?(p)->col:(pfcol(p),(p)->col))

B *bmk();
void brm();
B *bfind();
P *pdup();
P *pdupown();
P *poffline();
P *ponline();
B *bonline();
B *boffline();
void prm();
P *pset();
P *pbof();
P *peof();
int pisbof();
int piseof();
int piseol();
int pisbol();
int pisbow();
int piseow();
int pisblank();
int pisindent();
int pnext();
int pprev();
int pgetc();
P *pfwrd();
int prgetc();
P *pbkwd();
P *pgoto();
P *pfcol();
P *pbol();
P *peol();
P *pnextl();
P *pprevl();
P *pline();
P *pcolwse();
P *pcol();
P *pcoli();
void pbackws();
void pfill();
P *pfind();
P *pifind();
P *prfind();
P *prifind();
B *bcpy();
void pcoalesce();
void bdel();
P *binsb();
P *binsm();
P *binsc(P *, char);
P *binss();
B *bread();
B *bload();
B *bfind();
B *borphan();
int bsavefd();
int bsave();
char *parsens();
int brc();
char *brmem();
char *brs();
char *brvs();
B *bnext();
B *bprev();
char **getbufs();

#endif
