#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <pc.h>

#include "b.h"
#include "blocks.h"
#include "io.h"
#include "rc.h"
#include "ublock.h"
#include "zstr.h"
#include "bw.h"

/* Display modes */

extern int square, staen, skiptop;

int marking, mid;

P *getto(P *p, P *cur, P *top, int line)
{
	int dist=INT_MAX;
	int d;
	P *best=0;

	if (!p)
	{
		if (d=(line-cur->line>=0 ? line-cur->line : cur->line-line), d<dist)
			dist=d, best=cur;
		if (d=(line-top->line>=0 ? line-top->line : top->line-line), d<dist)
			dist=d, best=top;
		p=pdup(best);
		pbol (p);
	}
	while (line>p->line)
		if (!pnextl(p))
			break;
	if (line<p->line)
	{
		while (line<p->line)
			pprevl (p);
		pbol (p);
	}
	return p;
}

/* Scroll window to follow cursor */

void bwfllw(BW *w)
{
	P *newtop;

	if (w->cursor->line<w->top->line)
	{
		newtop=pdup(w->cursor);
		pbol (newtop);
		if (mid)
			if (newtop->line>=w->h/2)
				pline (newtop, newtop->line-w->h/2);
			else
				pset (newtop, newtop->b->bof);
		if (w->top->line-newtop->line<w->h)
			msetI (w->t->t->updtab+skiptop, -1, w->t->t->li-skiptop);
		else
			msetI (w->t->t->updtab+w->y, 1, w->h);
		pset (w->top, newtop);
		prm (newtop);
	}
	else if (w->cursor->line>=w->top->line+w->h)
	{
		newtop=pdup(w->top);
		if (mid)
			newtop=getto(NULL, w->cursor, w->top, w->cursor->line-w->h/2);
		else
			newtop=getto(NULL, w->cursor, w->top, w->cursor->line-(w->h-1));
		if (newtop->line-w->top->line<w->h)
			msetI (w->t->t->updtab+skiptop, -1, w->t->t->li-skiptop);
		else
			msetI (w->t->t->updtab+w->y, 1, w->h);
		pset (w->top, newtop);
		prm (newtop);
	}
/* Adjust column */
	if (w->cursor->xcol<w->offset)
	{
		w->offset=w->cursor->xcol;
		msetI (w->t->t->updtab+w->y, 1, w->h);
	}
	else if (w->cursor->xcol>=w->offset+w->w)
	{
		w->offset=w->cursor->xcol-(w->w-1);
		msetI (w->t->t->updtab+w->y, 1, w->h);
	}
	return;
}

/* Scroll a buffer window after an insert occured.  'flg' is set to 1 if
 * the first line was split
 */

void bwins(BW *w, int l, int n, int flg)
{
	if (l+flg+n<w->top->line+w->h && l+flg>=w->top->line && l+flg<=w->b->eof->line)
		msetI (w->t->t->updtab+skiptop, -1, w->t->t->li-skiptop);
	if (l<w->top->line+w->h && l>=w->top->line)
		if (n>=w->h-(l-w->top->line))
			msetI (w->t->t->updtab+w->y+l-w->top->line, 1, w->h-(int) (l-w->top->line));
		else
			msetI (w->t->t->updtab+w->y+l-w->top->line, 1, (int) n+1);
	return;
}

/* Scroll current windows after a delete */

void bwdel(BW *w, int l, int n, int flg)
{
/* Update the line where the delete began */
	if (l<w->top->line+w->h && l>=w->top->line)
		w->t->t->updtab[w->y+l-w->top->line]=1;
/* Update the line where the delete ended */
	if (l+n<w->top->line+w->h && l+n>=w->top->line)
		w->t->t->updtab[w->y+l+n-w->top->line]=1;
	if (l<w->top->line+w->h && (l+n>=w->top->line+w->h || (l+n==w->b->eof->line && w->b->eof->line>=w->top->line+w->h)))
		if (l>=w->top->line)
/* Update window from l to end */
			msetI (w->t->t->updtab+w->y+l-w->top->line, 1, w->h-(int) (l-w->top->line));
		else
/* Update entire window */
			msetI (w->t->t->updtab+w->y, 1, w->h);
	else if (l<w->top->line+w->h && l+n==w->b->eof->line &&	w->b->eof->line<w->top->line+w->h)
		if (l>=w->top->line)
/* Update window from l to end of file */
			msetI (w->t->t->updtab+w->y+l-w->top->line, 1, (int) n);
		else
/* Update from beginning of window to end of file */
			msetI (w->t->t->updtab+w->y, 1, (int) (w->b->eof->line-w->top->line));
	else if (l+n<w->top->line+w->h && l+n>w->top->line && l+n<w->b->eof->line)
		msetI (w->t->t->updtab+skiptop, -1, w->t->t->li-skiptop);
	return;
}

/* Update a single line */

static int lgen(SCRN *t, int y, int x, int w, P *p, int scr, int from, int to)
/* int w;            Window */
/* P *p;             Buffer pointer */
/* int scr;          Starting column to display */
/* int from, to;     Range for marked block */
{
	int ox=x;
	int col=0;
	int byte=p->byte;
	char *bp;                 /* Buffer pointer, 0 if not set */
	int amnt;                 /* Amount left in this segment of the buffer */
	int c, ta;
	unsigned char bc;

/* Initialize bp and amnt from p */
	if (p->ofst>=p->hdr->hole)
	{
		bp=p->ptr+p->hdr->ehole+p->ofst-p->hdr->hole;
		amnt=PGSIZE-p->hdr->ehole-(p->ofst-p->hdr->hole);
	}
	else
	{
		bp=p->ptr+p->ofst;
		amnt=p->hdr->hole-p->ofst;
	}
	if (col==scr)
		goto loop;
lp:                       /* Display next character */
	if (amnt)
		do
		{
			bc=*bp++;
			if (p->b->o.crlf && bc=='\r')
			{
				++byte;
				if (!--amnt)
				{
pppl:
					if (bp==p->ptr+PGSIZE)
					{
						if (pnext(p))
						{
							bp=p->ptr;
							amnt=p->hdr->hole;
						}
						else
							goto nnnl;
					}
					else
					{
						bp=p->ptr+p->hdr->ehole;
						amnt=PGSIZE-p->hdr->ehole;
						if (!amnt) goto pppl;
					}
				}
				if (*bp=='\n')
				{
					++bp;
					++byte;
					++amnt;
					goto eobl;
				}
nnnl:
				--byte;
				++amnt;
			}
			if (square)
				if (bc=='\t')
				{
					int tcol=col+p->b->o.tab-col%p->b->o.tab;

					if (tcol>from && tcol<=to)
						c=BLINK;
					else
						c=NORMAL;
				}
				else if (col>=from && col<to)
					c=BLINK;
				else
					c=NORMAL;
			else if (byte>=from && byte<to)
				c=BLINK;
			else
				c=NORMAL;
			++byte;
			if (bc=='\t')
			{
				ta=p->b->o.tab-col%p->b->o.tab;
				if (ta+col>scr)
				{
					ta-=scr-col;
					goto dota;
				}
				if ((col+=ta)==scr)
				{
					--amnt;
					goto loop;
				}
			}
			else if (bc=='\n')
				goto eobl;
			else if (++col==scr)
			{
				--amnt;
				goto loop;
			}
		} while (--amnt);
	if (bp==p->ptr+PGSIZE)
	{
		if (pnext(p))
		{
			bp=p->ptr;
			amnt=p->hdr->hole;
			goto lp;
		}
	}
	else
	{
		bp=p->ptr+p->hdr->ehole;
		amnt=PGSIZE-p->hdr->ehole;
		goto lp;
	}
	goto eof;
loop:                     /* Display next character */
	if (amnt)
		do
		{
			bc=*bp++;
			if (p->b->o.crlf && bc=='\r')
			{
				++byte;
				if (!--amnt)
				{
ppl:
					if (bp==p->ptr+PGSIZE)
					{
						if (pnext(p))
						{
							bp=p->ptr;
							amnt=p->hdr->hole;
						}
						else
							goto nnl;
					}
					else
					{
						bp=p->ptr+p->hdr->ehole;
						amnt=PGSIZE-p->hdr->ehole;
						if (!amnt)
							goto ppl;
					}
				}
				if (*bp=='\n')
				{
					++bp;
					++byte;
					++amnt;
					goto eobl;
				}
nnl:
				--byte;
				++amnt;
			}
			if (square)
				if (bc=='\t')
				{
					int tcol=scr+x-ox+p->b->o.tab-(scr+x-ox)%p->b->o.tab;

					if (tcol>from && tcol<=to)
						c=BLINK;
					else
						c=NORMAL;
				}
				else if (scr+x-ox>=from && scr+x-ox<to)
					c=BLINK;
				else
					c=NORMAL;
			else if (byte>=from && byte<to)
				c=BLINK;
			else
				c=NORMAL;
			++byte;
			if (bc=='\t')
			{
				ta=p->b->o.tab-((x-ox+scr)%p->b->o.tab);
dota:
				do
				{
					ScreenPutChar (' ', color[c], x, y);
					if (++x==w)
						goto eosl;
				} while (--ta);
			}
			else if (bc=='\n')
				goto eobl;
			else
			{
				ScreenPutChar (bc, color[c], x, y);
				if (++x==w)
					goto eosl;
			}
		} while (--amnt);
	if (bp==p->ptr+PGSIZE)
	{
		if (pnext(p))
		{
			bp=p->ptr;
			amnt=p->hdr->hole;
			goto loop;
		}
	}
	else
	{
		bp=p->ptr+p->hdr->ehole;
		amnt=PGSIZE-p->hdr->ehole;
		goto loop;
	}
	goto eof;
eobl:                     /* End of buffer line found.  Erase to end of screen line */
	++p->line;
eof:
	if (x!=w)
		eraeol (t, x, y);
/* Set p to bp/amnt */
	if (bp-p->ptr<=p->hdr->hole)
		p->ofst=bp-p->ptr;
	else
		p->ofst=bp-p->ptr-(p->hdr->ehole-p->hdr->hole);
	p->byte=byte;
	return 0;
eosl:
	if (bp-p->ptr<=p->hdr->hole)
		p->ofst=bp-p->ptr;
	else
		p->ofst=bp-p->ptr-(p->hdr->ehole-p->hdr->hole);
	p->byte=byte;
	pnextl (p);
	return 0;
}

void gennum(BW *w, int y)
{
	char buf[12];
	int z;
	int lin=w->top->line+y-w->y;

	if (lin<=w->b->eof->line)
		sprintf (buf, "%5d ", w->top->line+y-w->y+1);
	else
		zcpy (buf, "      ");
	for (z=0; buf[z]; ++z)
		ScreenPutChar (buf[z], color[NORMAL], z, y);
	return;
}

void bwgen(BW *w, int linums)
{
	P *p=0;
	P *q=pdup(w->cursor);
	int bot=w->h+w->y;
	int y;
	int dosquare=0;
	int from, to;
	int fromline, toline;
	SCRN *t=w->t->t;
	fromline=toline=from=to=0;

	if (markv(0) && markk->b==w->b)
		if (square)
		{
			from=markb->xcol, to=markk->xcol, dosquare=1;
			fromline=markb->line;
			toline=markk->line;
		}
		else
			from=markb->byte, to=markk->byte;
	else if (marking && markb && markb->b==w->b && w->cursor->byte!= markb->byte && !from)
		if (square)
		{
			from=Imin(w->cursor->xcol, markb->xcol),
			to=Imax(w->cursor->xcol, markb->xcol);
			fromline=Imin(w->cursor->line, markb->line);
			toline=Imax(w->cursor->line, markb->line);
			dosquare=1;
		}
	else
		from=Imin(w->cursor->byte, markb->byte),
	to=Imax(w->cursor->byte, markb->byte);
	if (marking)
		msetI (t->updtab+w->y, 1, w->h);
	for (y=w->cursor->line-w->top->line+w->y; y!=bot; ++y)
	{
		if (linums)
			gennum (w, y);
		if (t->updtab[y])
		{
			p=getto(p, w->cursor, w->top, w->top->line+y-w->y);
			if (dosquare)
				if (w->top->line+y-w->y>=fromline && w->top->line+y-w->y<=toline)
					t->updtab[y]=lgen(t, y, w->x, w->x+w->w, p, w->offset, from, to);
				else
					t->updtab[y]=lgen(t, y, w->x, w->x+w->w, p, w->offset, 0, 0);
			else
				t->updtab[y]=lgen(t, y, w->x, w->x+w->w, p, w->offset, from, to);
		}
	}
	for (y=w->y; y!=w->y+w->cursor->line-w->top->line; ++y)
	{
		if (linums)
			gennum (w, y);
		if (t->updtab[y])
		{
			p=getto(p, w->cursor, w->top, w->top->line+y-w->y);
			if (dosquare)
				if (w->top->line+y-w->y>=fromline && w->top->line+y-w->y<=toline)
					t->updtab[y]=lgen(t, y, w->x, w->x+w->w, p, w->offset, from, to);
				else
					t->updtab[y]=lgen(t, y, w->x, w->x+w->w, p, w->offset, 0, 0);
			else
				t->updtab[y]=lgen(t, y, w->x, w->x+w->w, p, w->offset, from, to);
		}
	}
	prm (q);
	if (p)
		prm (p);
	return;
}

void bwmove(BW *w, int x, int y)
{
	w->x=x;
	w->y=y;
	return;
}

void bwresz(BW *w, int wi, int he)
{
	if (he>w->h && w->y!=-1)
		msetI (w->t->t->updtab+w->y+w->h, 1, he-w->h);
	w->w=wi;
	w->h=he;
	return;
}

BW *bwmk(W *window, B *b, int prompt)
{
	BW *w=(BW *) malloc(sizeof(BW));

	w->parent=window;
	w->b=b;
	if (prompt || (!window->y && staen))
		w->y=window->y, w->h=window->h;
	else
		w->y=window->y+1, w->h=window->h-1;
	if (b->oldcur)
	{
		w->top=b->oldtop, b->oldtop=0, w->top->owner=0;
		w->cursor=b->oldcur, b->oldcur=0, w->cursor->owner=0;
	}
	else
	{
		w->top=pdup(b->bof);
		w->cursor=pdup(b->bof);
	}
	w->t=window->t;
	w->object=NULL;
	w->offset=0;
	w->o=w->b->o;
	if (w->o.linums)
		w->x=window->x+LINCOLS, w->w=window->w-LINCOLS;
	else
		w->x=window->x, w->w=window->w;
	if (window==window->main)
	{
		free (window->kbd);
		window->kbd=mkkbd(getcontext(w->o.context));
	}
	w->top->xcol=0;
	w->cursor->xcol=0;
	return w;
}

void bwrm(BW *w)
{
	prm (w->top);
	prm (w->cursor);
	brm (w->b);
	free (w);
	return;
}

int ustat(BW *bw)
{
	static char buf[80];
	unsigned c=brc(bw->cursor);

	if (c==INT_MAX)
		sprintf (buf, "** Line %d  Col %d  Offset %d(0x%x) **", bw->cursor->line+1, piscol(bw->cursor)+1, bw->cursor->byte, bw->cursor->byte);
	else
		sprintf (buf, "** Line %d  Col %d  Offset %d(0x%x)  Ascii %d(0%o/0x%X) **", bw->cursor->line+1, piscol(bw->cursor)+1, bw->cursor->byte, bw->cursor->byte, 255&c, 255&c, 255&c);
	msgnw (bw, buf);
	return 0;
}

int ucrawlr(BW *bw)
{
	int amnt=bw->w/2;

	pcol (bw->cursor, bw->cursor->xcol+amnt);
	bw->cursor->xcol+=amnt;
	bw->offset+=amnt;
	updall ();
	return 0;
}

int ucrawll(BW *bw)
{
	int amnt=bw->w/2;
	int curamnt=bw->w/2;

	if (amnt>bw->offset)
		amnt=bw->offset, curamnt=bw->offset;
	if (!bw->offset)
		curamnt=bw->cursor->xcol;
	if (!curamnt)
		return -1;
	pcol (bw->cursor, bw->cursor->xcol-curamnt);
	bw->cursor->xcol-=curamnt;
	bw->offset-=amnt;
	updall ();
	return 0;
}

void orphit(BW *bw)
{
	++bw->b->count;
	bw->b->orphan=1;
	pdupown (bw->cursor, &bw->b->oldcur);
	pdupown (bw->top, &bw->b->oldtop);
	return;
}
