#include <pc.h>

#include "blocks.h"
#include "io.h"
#include "w.h"
#include "zstr.h"
#include "help.h"

/* The loaded help screen */

extern int skiptop;

int hlpsiz, hlpbsz, hlplns, helpcursor;
char *hlptxt=0;
char **help_names;
struct help **help_structs;
struct help *first_help;

int get_help(char *name)
{
	int x;

	for (x=0; help_structs[x]; ++x)
		if (!zcmp(help_structs[x]->name, name))
			break;
	if (help_structs[x])
		return x;
	else
		return -1;
}

/* Display help text */

void dsphlp(SCREEN *t)
{
	char *str=hlptxt;
	int y, x, c;
	int atr=NORMAL;

	for (y=skiptop; y!=t->wind; ++y)
	{
		if (t->t->updtab[y])
		{
			for (x=0; x!=t->w-1; ++x)
				if (*str=='\n' || !*str)
				{
					eraeol (t->t, x, y);
					break;
				}
				else
				{
					if (*str=='\\')
						switch(*++str)
						{
							case 'i': case 'I':
								atr^=INVERSE;
								++str;
								--x;
								goto cont;
							case 'u': case 'U':
								atr^=UNDERLINE;
								++str;
								--x;
								goto cont;
							case 'd': case 'D':
								atr^=DIM;
								++str;
								--x;
								goto cont;
							case 'b': case 'B':
								atr^=BOLD;
								++str;
								--x;
								goto cont;
							case 'f': case 'F':
								atr^=BLINK;
								++str;
								--x;
								goto cont;
							case 0:
								--x;
								goto cont;
							default:
								c=(unsigned char) *str++;
					}
					else
						c=(unsigned char) *str++;
				ScreenPutChar (c, color[atr], x, y);
cont:
			}
			atr=NORMAL;
			t->t->updtab[y]=0;
		}
		while (*str && *str!='\n')
			++str;
		if (*str=='\n')
			++str;
	}
	return;
}

/* Create the help window */

int helpon(SCREEN *t)
{
	struct help *h=help_structs[helpcursor];

	hlptxt=h->hlptxt;
	hlpsiz=h->hlpsiz;
	hlpbsz=h->hlpbsz;
	hlplns=h->hlplns;
	if (!hlptxt)
		return -1;
	t->wind=hlplns+skiptop;
	if (t->h-t->wind<FITHEIGHT)
		t->wind=t->h-FITHEIGHT;
	if (t->wind<0)
	{
		t->wind=skiptop;
		return -1;
	}
	wfit (t);
	msetI (t->t->updtab+skiptop, 1, t->wind);
	return 0;
}

/* Eliminate the help window */

void helpoff(SCREEN *t)
{
	t->wind=skiptop;
	wfit (t);
	return;
}

/* Toggle help on/off */

int uhelp(BASE *base)
{
	W *w=base->parent;
	int h;

	if (w->huh && (h=get_help(w->huh))>-1)
	{
		if (w->t->wind!=skiptop)
			helpoff (w->t);
		helpcursor=h;
		return	helpon(w->t);
	}
	else if (w->t->wind==skiptop)
		return helpon(w->t);
	else
	{
		helpoff (w->t);
		return 0;
	}
}

/* Goto next/prev help screen */

int uhnext(BASE *base)
{
	W *w=base->parent;

	if (help_structs[helpcursor+1])
	{
		if (w->t->wind!=skiptop)
			helpoff(w->t);
		++helpcursor;
		return helpon(w->t);
	}
	else return -1;
}

int uhprev(BASE *base)
{
	W *w=base->parent;

	if (helpcursor)
	{
		if (w->t->wind!=skiptop)
			helpoff(w->t);
		--helpcursor;
		return helpon(w->t);
	}
	else return -1;
}
