/* Screen, keyboard and path functions */

#ifndef _Iio
#define _Iio 1

#define KEYS 256

/* Character attribute bits */

#define NORMAL 0
#define INVERSE 1
#define UNDERLINE 2
#define BOLD 4
#define BLINK 8
#define DIM 16

extern int color[];

typedef struct scrn SCRN;
typedef struct kmap KMAP;
typedef struct kbd KBD;
typedef struct key KEY;

struct hentry
{
	int next;
	int loc;
};

struct scrn
{
	int li;
	int co;
	int *updtab;
	int *sary;
};

/* A key binding */

struct key
{
	int k;                    /* Flag: 0=binding, 1=submap */
	union
	{
		void *bind;           /* What key is bound to */
		KMAP *submap;         /* Sub KMAP address (for prefix keys) */
	} value;
};

/* A map of keycode to command/sub-map bindings */

struct kmap
{
	KEY keys[KEYS];           /* KEYs */
};

/** A keyboard handler **/

struct kbd
{
	KMAP *curmap;             /* Current keymap */
	KMAP *topmap;             /* Top-level keymap */
	int seq[16];              /* Current sequence of keys */
	int x;                    /* What we're up to */
};

/* KMAP *mkkmap(void);
 * Create an empty keymap
 */
KMAP *mkkmap();

/* void rmkmap(KMAP *kmap);
 * Free a key map
 */
void rmkmap();

/* int kadd(KMAP *kmap,char *seq,void *bind);
 * Add a key sequence binding to a key map
 *
 * Returns 0 for success
 *        -1 for for invalid key sequence
 *
 * A valid key sequence is one or more keys seperated with spaces.  A key
 * is a single character or one of the following strings:
 *
 *    ^?                      127 (DEL)
 *
 *    ^@   -   ^_                  Control characters
 *
 *    SP                           32 (space character)
 *
 *    UP, DOWN, LEFT, RIGHT,
 *    F0 - F10, DEL, INS, HOME,
 *    END, PGUP, PGDN              termcap special characters
 *
 * In addition, the last key of a key sequence may be replaced with
 * a range-fill of the form: <KEY> TO <KEY>
 *
 * So for example, if the sequence: ^K A TO Z
 * is speicified, then the key sequences
 * ^K A, ^K B, ^K C, ... ^K Z are all bound.
 */
int kadd();

/* void kcpy(KMAP *dest,KMAP *src);
 * Copy all of the entries in the 'src' keymap into the 'dest' keymap
 */
void kcpy();

/* int kdel(KMAP *kmap,char *seq);
 * Delete a binding from a keymap
 *
 * Returns 0 for success
 *        -1 if the given key sequence was invalid
 *         1 if the given key sequence did not exist
 */
int kdel();

/* KBD *mkkbd(KMAP *kmap);
 * Create a keyboard handler which uses the given keymap
 */
KBD *mkkbd();

/* void *dokey(KBD *kbd,int k);
 * Handle a key for a KBD:
 * Returns 0 for invalid or prefix keys
 * Returns binding for a completed key sequence
 */
void *dokey();

/* char *namprt(char *path);
 * Return name part of a path.  There is no name if the last character
 * in the path is '/'.
 */
char *namprt();

/* char *dirprt(char *path);
 * Return directory and drive part of a path.  I.E., everything to the
 * left of the name part.
 */
char *dirprt();

/* char *begprt(char *path);
 * Return the beginning part of a path.
 *
 * The beginning part of "/hello/there" is "/hello/"
 * The beginning part of "/hello/" is "/"
 * The beginning part of "/" is "/"
 */
char *begprt();

/* char *endprt(char *path);
 * Return the ending part of a path.
 *
 * The ending part of "/hello/there" is "there"
 * The ending part of "/hello/" is "hello/"
 * The ending part of "/" is ""
 */
char *endprt();

/* int rmatch(char *pattern,char *string);
 * Return true if string matches pattern
 *
 * Pattern is:
 *     *                 matches 0 or more charcters
 *     ?                 matches any 1 character
 *     [...]             matches 1 character in set
 *     [^...]            matches 1 character not in set
 *     [[]               matches [
 *     [*]               matches *
 *     [?]               matches ?
 *     any other         matches self
 *
 *  Ranges of characters may be specified in sets with 'A-B'
 *  '-' may be specified in sets by placing it at the ends
 *  '[' may be specified in sets by placing it first 
 */
int rmatch();

/* char **rexpnd(char *path,char *pattern);
 * Generate array (see va.h) of file names from directory in 'path'
 * which match the pattern 'pattern'
 */
char **rexpnd();

SCRN *nopen();
void nclose();
void eraeol();
int getch();
void bell();
char *joesep();
int isreg();
int chddir();

#endif
