#include <stdio.h>
#include <limits.h>

#include "b.h"
#include "main.h"
#include "misc.h"
#include "tw.h"
#include "ufile.h"
#include "vs.h"
#include "zstr.h"
#include "uerror.h"

/* Error database */

typedef struct error ERROR;

B *errbuf=0;                  /* Buffer with error messages */

struct error
{
	LINK(ERROR) link;         /* Linked list of errors */
	int line;                 /* Target line number */
	int org;                  /* Original target line number */
	char *file;               /* Target file name */
	int src;                  /* Error-file line number */
	char *msg;                /* The message */
}

errors=
{
	{
		&errors, &errors
	}
};

ERROR *errptr=&errors;        /* Current error row */

/* Insert and delete notices */

void inserr(char *name, int where, int n, int bol)
{
	ERROR *e;

	if (name)
		for (e=errors.link.next; e!=&errors; e=e->link.next)
			if (!zcmp(e->file, name))
				if (e->line>where)
					e->line+=n;
				else if (e->line==where && bol)
					e->line+=n;
	return;
}

void delerr(char *name, int where, int n)
{
	ERROR *e;

	if (name)
		for (e=errors.link.next; e!=&errors; e=e->link.next)
			if (!zcmp(e->file, name))
				if (e->line>where+n)
					e->line-=n;
				else if (e->line>where)
					e->line=where;
	return;
}

/* Abort notice */

void abrerr(char *name)
{
	ERROR *e;

	if (name)
		for (e=errors.link.next;e!=&errors;e=e->link.next)
			if (!zcmp(e->file, name)) e->line=e->org;
	return;
}

/* Save notice */

void saverr(char *name)
{
	ERROR *e;

	if (name)
		for (e=errors.link.next; e!=&errors; e=e->link.next)
			if (!zcmp(e->file, name))
				e->org=e->line;
	return;
}

/* Pool of free error nodes */
ERROR errnodes=
{
	{
		&errnodes, &errnodes
	}
};

/* Free an error node */

void freeerr(ERROR *n)
{
	vsrm (n->file);
	vsrm (n->msg);
	(void) enquef (ERROR, link, &errnodes, n);
	return;
}

/* Free all errors */

void freeall()
{
	while (!qempty(ERROR, link, &errors))
		freeerr (deque(ERROR, link, errors.link.next));
	errptr=&errors;
	return;
}

/* Parse error messages into database */

int parseit(char *s, int row)
{
	int x, y;
	char *name=0;
	int line=-1;
	ERROR *err;

/* Skip to first word */
	for (x=0; s[x] && !(crest(s[x]) || s[x]=='.' || s[x]=='/'); ++x);
/* Skip to end of first word */
	for (y=x; crest(s[y]) || s[y]=='.' || s[y]=='/'; ++y);
/* Save file name */
	if (x!=y)
		name=vsncpy(NULL, 0, s+x, y-x);
/* Skip to first number */
	for (x=y; s[x] && (s[x]<'0' || s[x]>'9'); ++x);
/* Skip to end of first number */
	for (y=x; s[y]>='0' && s[y]<='9'; ++y);
/* Save line number */
	if (x!=y)
		sscanf (s+x, "%d", &line);
	if (line!=-1)
		--line;
	if (name)
		if (line!=-1)
		{
/* We have an error */
			err=(ERROR *) alitem(&errnodes, sizeof(ERROR));
			err->file=name;
			err->org=err->line=line;
			err->src=row;
			err->msg=vsncpy(NULL, 0, sc("\\i"));
			err->msg=vsncpy(sv(err->msg), sv(s));
			(void) enqueb (ERROR, link, &errors, err);
			return 1;
		}
		else vsrm (name);
	return 0;
}

/* Parse the error output contained in a buffer */

int parserr(B *b)
{
	P *p=pdup(b->bof);
	P *q=pdup(p);
	int nerrs=0;

	freeall ();
	do
	{
		char *s;

		pset (q, p);
		peol (p);
		s=brvs(q, (int) (p->byte-q->byte));
		if (s)
		{
			nerrs+=parseit(s, q->line);
			vsrm (s);
		}
	} while (pgetc(p)!=INT_MAX);
	prm (p);
	prm (q);
	return nerrs;
}

int uparserr(BW *bw)
{
	errbuf=bw->b;
	freeall ();
	sprintf (msgbuf, "Parsed %d lines", parserr(bw->b));
	msgnw (bw, msgbuf);
	return 0;
}

int unxterr(BW *bw)
{
	int omid;

	if (errptr->link.next==&errors)
	{
		msgnw (bw, "No more errors");
		return -1;
	}
	errptr=errptr->link.next;
	if (!bw->b->name || zcmp(errptr->file, bw->b->name))
	{
		if (doedit(bw, vsdup(errptr->file), NULL, NULL))
			return -1;
		bw=(BW *) maint->curwin->object;
	}
	omid=mid;
	mid=1;
	pline (bw->cursor, errptr->line);
	setline (errbuf, errptr->src);
	dofollows ();
	mid=omid;
	bw->cursor->xcol=piscol(bw->cursor);
	msgnw (bw, errptr->msg);
	return 0;
}

int uprverr(BW *bw)
{
	int omid;

	if (errptr->link.prev==&errors)
	{
		msgnw (bw, "No more errors");
		return -1;
	}
	errptr=errptr->link.prev;
	if (!bw->b->name || zcmp(errptr->file, bw->b->name))
	{
		if (doedit(bw, vsdup(errptr->file), NULL, NULL))
			return -1;
		bw=(BW *) maint->curwin->object;
	}
	omid=mid;
	mid=1;
	pline (bw->cursor, errptr->line);
	setline (errbuf, errptr->src);
	dofollows ();
	mid=omid;
	bw->cursor->xcol=piscol(bw->cursor);
	msgnw (bw, errptr->msg);
	return 0;
}
