#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <fcntl.h>
#include <pc.h>

#include "b.h"
#include "blocks.h"
#include "bw.h"
#include "io.h"
#include "macro.h"
#include "main.h"
#include "menu.h"
#include "pw.h"
#include "qw.h"
#include "tab.h"
#include "tw.h"
#include "ublock.h"
#include "uerror.h"
#include "va.h"
#include "vs.h"
#include "zstr.h"
#include "ufile.h"

extern int orphan, skiptop;

int nobackups, exask;

char *backpath=0;             /* Place to store backup files */
static B *filehist=0;         /* History of file names */

struct savereq
{
	int (*callback)();
	char *name;
};

/* Ending message generator */

void genexmsg(BW *bw, int saved, char *name)
{
	char *s;

	if (bw->b->name && bw->b->name[0])
		s=bw->b->name;
	else
		s="(Unnamed)";
	if (name)
		if (saved)
			sprintf (msgbuf, "File %s saved", name);
		else
			sprintf (msgbuf, "File %s not saved", name);
	else if (bw->b->changed && bw->b->count==1)
			sprintf (msgbuf, "File %s not saved", s);
		else if (saved)
			sprintf (msgbuf, "File %s saved", s);
	else
		sprintf (msgbuf, "File %s not changed so no update needed", s);
	msgnw (bw, msgbuf);
	if (!exmsg)
		if (bw->b->changed && bw->b->count==1)
		{
			exmsg=vsncpy(NULL, 0, sc("File "));
			exmsg=vsncpy(sv(exmsg), sz(s));
			exmsg=vsncpy(sv(exmsg), sc(" not saved."));
		}
		else if (saved)
		{
			exmsg=vsncpy(NULL, 0, sc("File "));
			exmsg=vsncpy(sv(exmsg), sz(s));
			exmsg=vsncpy(sv(exmsg), sc(" saved."));
		}
		else
		{
			exmsg=vsncpy(NULL, 0, sc("File "));
			exmsg=vsncpy(sv(exmsg), sz(s));
			exmsg=vsncpy(sv(exmsg), sc(" not changed so no update needed."));
		}
	return;
}

/* Write highlighted block to a file */

int ublksave(BW *bw)
{
	if (markb && markk && markb->b==markk->b && (markk->byte-markb->byte)>0 && (!square || piscol(markk)>piscol(markb)))
	{
		if (wmkpw(bw, "Name of file to write (^C to abort): ", &filehist, dowrite, "Names", NULL, cmplt, NULL, NULL))
			return 0;
		else
			return -1;
	}
	else
		return usave(bw);
}

/* Shell escape */

int ushell(BW *bw)
{
	char command[256];

	ScreenSetCursor (bw->parent->t->t->li-1, 0);
	eraeol (bw->parent->t->t, 0, bw->parent->t->t->li-1);
	puts ("\nType \"exit\" to continue editing ...\n");
	sprintf (command, "%s /K PROMPT [JOE] %s", getenv("COMSPEC"), getenv("PROMPT"));
	system (command);
	msetI (bw->parent->t->t->updtab+skiptop, -1, bw->parent->t->t->li-skiptop);
	return 0;
}

/* Start documentation system */

int uinfo(BW *bw)
{
	system ("info >& nul:");
	msetI (bw->parent->t->t->updtab+skiptop, -1, bw->parent->t->t->li-skiptop);
	return 0;
}

/* Copy a file */

int cp(char *from, char *to)
{
	int f, g, amnt;
	void* buffer;

	if ((f=open(from, O_RDONLY))>0)
		if ((g=open(to, O_CREAT | O_WRONLY))>0)
			if ((buffer=malloc(65536)))
			{
				while ((amnt=read(f, buffer, 65536))>0)
					if (amnt!=write(g, buffer, amnt))
						break;
			}
			else
			{
				close (f);
				close (g);
				return -1;
			}
		else
		{
			close (f);
			return -1;
		}
	else
		return -1;
	free (buffer);
	if (amnt)
		return -1;
	else
		return 0;
}

/* Make backup file if it needs to be made
 * Returns 0 if backup file was made or didn't need to be made
 * Returns 1 for error
 */

static int backup(BW *bw)
{
	if (!bw->b->backup && !nobackups && bw->b->name && bw->b->name[0])
	{
		char name[128];
		int x;

		if (backpath)
			sprintf (name, "%s/%s", backpath, namprt(bw->b->name));
		else
			sprintf (name, "%s", bw->b->name);
		for (x=zlen(name); name[--x]!='.'; )
			if (name[x]=='/' || name[x]=='\\' || name[x]==':' || x==0)
			{
				x=zlen(name);
				break;
			}
		zcpy (name+x, ".bak");
/* Copy original file to backup file */
		if (cp(bw->b->name, name))
			return 1;
		else
		{
			bw->b->backup=1;
			return 0;
		}
	}
	else
		return 0;
}

/* Write file */

static int saver(BW *bw, int c, struct savereq *req, int *notify)
{
	int (*callback)();
	int fl;

	callback=req->callback;
	if (c=='n' || c=='N')
	{
		vsrm (req->name);
		free (req);
		if (notify)
			*notify=1;
		msgnw (bw, "Couldn't make backup file... file not saved");
		if (callback)
			return callback(bw, -1);
		else
			return -1;
	}
	if (c!='y' && c!='Y')
		if (mkqw(bw, sc("Could not make backup file.  Save anyway (y, n, ^C)? "), saver, NULL, req, notify))
			return 0;
	else
	{
		if (notify)
			*notify=1;
		return -1;
	}
	if (notify)
		*notify=1;
	if (bw->b->er==-1 && bw->o.msnew)
		exemac (bw->o.msnew), bw->b->er=-3;
	if (bw->b->er== 0 && bw->o.msold)
		exemac (bw->o.msold);
	if ((fl=bsave(bw->b->bof, req->name, bw->b->eof->byte)))
	{
		msgnw (bw, msgs[fl+5]);
		vsrm (req->name);
		free (req);
		if (callback)
			return callback(bw, -1);
		else
			return -1;
	}
	else
	{
		if (!bw->b->name)
			bw->b->name=joesep(zdup(req->name));
		if (!zcmp(bw->b->name, req->name))
		{
			bw->b->changed=0;
			saverr (bw->b->name);
		}
		genexmsg (bw, 1, req->name);
		vsrm (req->name);
		free (req);
		if (callback)
			return callback(bw, 0);
		else
			return 0;
	}
}

static int dosave(BW *bw, char *s, int (*callback)(), int *notify)
{
	struct savereq *req=(struct savereq *) malloc(sizeof(struct savereq));

	req->name=s;
	req->callback=callback;
	if (backup(bw))
		return saver(bw, 0, req, notify);
	else
		return saver(bw, 'y', req, notify);
}

static int dosave2(BW *bw, int c, char *s, int *notify)
{
	if (c=='y' || c=='Y')
		return dosave(bw, s, NULL, notify);
	else if (c=='n' || c=='N')
	{
		if (notify)
			*notify=1;
		genexmsg (bw, 0, s);
		vsrm (s);
		return -1;
	}
	else
		if (mkqw(bw, sc("File exists.  Overwrite (y, n, ^C)? "), dosave2, NULL, s, notify))
			return 0;
		else
			return -1;
}

static int dosave1(BW *bw, char *s, void *object, int *notify)
{
	int f;

	if (s[0]!='!' && !(s[0]=='>' && s[1]=='>') && (!bw->b->name || zcmp(s, bw->b->name)))
	{
		f=open(s, O_RDONLY);
		if (f!=-1)
		{
			close (f);
			return dosave2(bw, 0, s, notify);
		}
	}
	return dosave(bw, s, object, notify);
}

int usave(BW *bw)
{
	BW *pbw=wmkpw(bw, "Name of file to save (^C to abort): ", &filehist, dosave1, "Names", NULL, cmplt, NULL, NULL);
	if (pbw && bw->b->name)
	{
		binss (pbw->cursor, bw->b->name);
		pset (pbw->cursor, pbw->b->eof);
		pbw->cursor->xcol=piscol(pbw->cursor);
	}
	if (pbw)
		return 0;
	else
		return -1;
}

/* Load file to edit */

int doedit(BW *bw, char *s, void *obj, int *notify)
{
	int ret=0;
	int er;
	void *object;
	W *w;
	B *b;

	if (notify)
		*notify=1;
	b=bfind(s);
	er=error;
	if (bw->b->count==1 && (bw->b->changed || bw->b->name))
		if (orphan)
			orphit (bw);
		else
		{
			if (uduptw(bw))
			{
				brm (b);
				return -1;
			}
			bw=(BW *) maint->curwin->object;
		}
	if (er)
	{
		msgnwt (bw, msgs[er+5]);
		if (er!=-1)
			ret=-1;
	}
	object=bw->object;
	w=bw->parent;
	bwrm (bw);
	w->object=(void *) (bw=bwmk(w, b, 0));
	wredraw (bw->parent);
	bw->object=object;
	vsrm (s);
	if (er==-1 && bw->o.mnew)
		exemac (bw->o.mnew);
	if (er==0 && bw->o.mold)
		exemac (bw->o.mold);
	return ret;
}

int okrepl(BW *bw)
{
	if (bw->b->count==1 && bw->b->changed)
	{
		msgnw (bw, "Can't replace modified file");
		return -1;
	}
	else return 0;
}

int uedit(BW *bw)
{
	if (wmkpw(bw, "Name of file to edit (^C to abort): ", &filehist, doedit, "Names", NULL, cmplt, NULL, NULL))
		return 0;
	else
		return -1;
}

/* Load file into buffer: can result in an orphaned buffer */

int dorepl(BW *bw, char *s, void *obj, int *notify)
{
	void *object=bw->object;
	int ret=0;
	int er;
	W *w=bw->parent;
	B *b;

	if (notify)
		*notify=1;
	b=bfind(s);
	er=error;
	if (error)
	{
		msgnwt (bw, msgs[error+5]);
		if (error!=-1)
			ret=-1;
	}
	if (bw->b->count==1 && (bw->b->changed || bw->b->name))
		orphit (bw);
	bwrm (bw);
	w->object=(void *) (bw=bwmk(w, b, 0));
	wredraw (bw->parent);
	bw->object=object;
	vsrm (s);
	if (er==-1 && bw->o.mnew)
		exemac (bw->o.mnew);
	if (er==0 && bw->o.mold)
		exemac (bw->o.mold);
	return ret;
}

/* Switch to next buffer in window */

int unbuf(BW *bw)
{
	void *object=bw->object;
	W *w=bw->parent;
	B *b;

	b=bnext();
	if (b==bw->b)
		b=bnext();
	if (b==bw->b)
		return -1;
	if (!b->orphan)
		++b->count;
	else
		b->orphan=0;
	if (bw->b->count==1)
		orphit(bw);
	bwrm (bw);
	w->object=(void *) (bw=bwmk(w, b, 0));
	wredraw (bw->parent);
	bw->object=object;
	return 0;
}

int upbuf(BW *bw)
{
	void *object=bw->object;
	W *w=bw->parent;
	B *b;

	b=bprev();
	if (b==bw->b)
		b=bprev();
	if (b==bw->b)
		return -1;
	if (!b->orphan)
		++b->count;
	else
		b->orphan=0;
	if (bw->b->count==1)
		orphit (bw);
	bwrm (bw);
	w->object=(void *) (bw=bwmk(w, b, 0));
	wredraw (bw->parent);
	bw->object=object;
	return 0;
}

int uinsf(BW *bw)
{
	if (wmkpw(bw, "Name of file to insert (^C to abort): ", &filehist, doinsf, "Names", NULL, cmplt, NULL, NULL))
		return 0;
	else
		return -1;
}

/* Save and exit */

static int exdone(BW *bw, int flg)
{
	if (flg)
	{
		if (bw->b->name)
			free (bw->b->name);
		bw->b->name=0;
		return -1;
	}
	else
	{
		bw->b->changed=0;
		saverr (bw->b->name);
		return uabort(bw, INT_MAX);
	}
}

static int exdone1(BW *bw, int flg)
{
	if (flg)
		return -1;
	else
	{
		bw->b->changed=0;
		saverr (bw->b->name);
		return uabort(bw, INT_MAX);
	}
}

static int doex(BW *bw, char *s, void *object, int *notify)
{
	bw->b->name=joesep(zdup(s));
	return dosave(bw, s, exdone, notify);
}

int uexsve(BW *bw)
{
	if (!bw->b->changed)
	{
		uabort (bw, INT_MAX);
		return 0;
	}
	else if (bw->b->name && !exask)
	{
		return dosave(bw, vsncpy(NULL, 0, sz(bw->b->name)), exdone1, NULL);
	}
	else
	{
		BW *pbw=wmkpw(bw, "Name of file to save (^C to abort): ", &filehist, doex, "Names", NULL, cmplt, NULL, NULL);
		if (pbw && bw->b->name)
		{
			binss (pbw->cursor, bw->b->name);
			pset (pbw->cursor, pbw->b->eof);
			pbw->cursor->xcol=piscol(pbw->cursor);
		}
		if (pbw)
			return 0;
		else
			return -1;
	}
}

/* If buffer is modified, prompt for saving */

static int nask(BW *bw, int c, void *object, int *notify)
{
	if (c=='y' || c=='Y')
		if (bw->b->name)
			return dosave1(bw, vsncpy(NULL, 0, sz(bw->b->name)), object, notify);
	else
	{
		BW *pbw=wmkpw(bw, "Name of file to save (^C to abort): ", &filehist, dosave1, "Names", NULL, cmplt, object, notify);

		if (pbw)
			return 0;
		else
			return -1;
	}
	else if (c=='n' || c=='N')
	{
		genexmsg (bw, 0, NULL);
		if (notify)
			*notify=1;
		return 0;
	}
	else
		if (bw->b->count==1 && bw->b->changed)
		{
			if (mkqw(bw, sc("Save changes to this file (y, n, ^C)? "), nask, NULL, object, notify))
				return 0;
			else
				return -1;
		}
		else
		{
			if (notify)
				*notify=1;
			return 0;
		}
}

/* If buffer is modified, ask if it's ok to lose changes */

int dolose(BW *bw, int c, void *object, int *notify)
{
	W *w;

	if (notify)
		*notify=1;
	if (c!='y' && c!='Y')
		return -1;
	genexmsg (bw, 0, NULL);
	if (bw->b->count==1)
		bw->b->changed=0;
	object=bw->object;
	w=bw->parent;
	bwrm (bw);
	w->object=(void *) (bw=bwmk(w, bfind(""), 0));
	wredraw (bw->parent);
	bw->object=object;
	if (bw->o.mnew)
		exemac (bw->o.mnew);
	return 0;
}

int ulose(BW *bw)
{
	msgnw (bw, NULL);
	if (bw->b->count==1 && bw->b->changed)
		if (mkqw(bw, sc("Lose changes to this file (y, n, ^C)? "), dolose, NULL, NULL, NULL))
			return 0;
		else
			return -1;
	else
		return dolose(bw, 'y', NULL, NULL);
}

/* Buffer list */

int dobuf(MENU *m, int x, char **s)
{
	char *name;
	BW *bw=m->parent->win->object;
	int *notify=m->parent->notify;

	m->parent->notify=0;
	name=vsdup(s[x]);
	uabort (m, INT_MAX);
	return dorepl(bw, name, NULL, notify);
}

int abrtb(MENU *m, int x, char **s)
{
	varm (s);
	return -1;
}

int ubufed(BW *bw)
{
	char **s=getbufs();

	vasort (av(s));
	if (mkmenu(bw, s, dobuf, abrtb, NULL, 0, s, NULL))
		return 0;
	else
	{
		varm (s);
		return -1;
	}
}
