#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <fcntl.h>

#include "blocks.h"
#include "misc.h"
#include "vfile.h"

static VFILE vfiles=          /* Known vfiles */
{
	{
		&vfiles, &vfiles
	}
};

static VPAGE *freepages=0;    /* Linked list of free pages */
static VPAGE *htab[HTSIZE];   /* Hash table of page headers */
char *vbase;                  /* Data first entry in vheader refers to */
VPAGE **vheaders=0;           /* Array of header addresses */
static int vheadsz=0;         /* No. entries allocated to vheaders */

char *vlock(VFILE *vfile, int addr)
{
	VPAGE *vp, *pp;
	int x, y;
	int ofst=(addr&(PGSIZE-1));

	addr-=ofst;
	for (vp=htab[((addr>>LPGSIZE)+(int) vfile)&(HTSIZE-1)]; vp; vp=vp->next)
		if (vp->vfile==vfile && vp->addr==addr)
			return ++vp->count, vp->data+ofst;
	if (freepages)
	{
		vp=freepages;
		freepages=vp->next;
		goto gotit;
	}
	if ((vp=(VPAGE *) malloc(sizeof(VPAGE))))
	{
		vp->data=PGSIZE+(char *) malloc(2*PGSIZE);
		if (vp->data-PGSIZE)
		{
			if (!vheaders)
				vheaders=(VPAGE **) malloc(sizeof(VPAGE *)), vbase=vp->data, vheadsz=1;
			else if ((unsigned) vp->data<(unsigned) vbase)
			{
				VPAGE **t=vheaders;
				int amnt=((unsigned) vbase-(unsigned) vp->data)>>LPGSIZE;

				vheaders=(VPAGE **)malloc ((amnt+vheadsz)*sizeof(VPAGE *));
				mcpy (vheaders+amnt, t, vheadsz*sizeof(VPAGE *));
				vheadsz+=amnt;
				vbase=vp->data;
				free (t);
			}
			else if ((((unsigned) (vp->data+PGSIZE)-(unsigned) vbase)>>LPGSIZE)>vheadsz)
				vheaders=(VPAGE **) realloc(vheaders, (vheadsz=((((unsigned) (vp->data+PGSIZE)-(unsigned) vbase)>>LPGSIZE)))*sizeof(VPAGE *));
			vheader (vp->data)=vp;
			goto gotit;
		}
		free (vp);
		vp=0;
	}
	for (y=HTSIZE, x=(random()&(HTSIZE-1)); y; x=((x+1)&(HTSIZE-1)), --y)
		for (pp=(VPAGE *) (htab+x), vp=pp->next; vp; pp=vp, vp=vp->next)
			if (!vp->count && !vp->dirty)
			{
				pp->next=vp->next;
				goto gotit;
			}
	fprintf (stderr, "\nOut of memory!\n");
	fflush (stderr);
	exit (1);
gotit:
	vp->addr=addr;
	vp->vfile=vfile;
	vp->dirty=0;
	vp->count=1;
	vp->next=htab[((addr>>LPGSIZE)+(int) vfile)&(HTSIZE-1)];
	htab[((addr>>LPGSIZE)+(int) vfile)&(HTSIZE-1)]=vp;
	mset (vp->data, 0, PGSIZE);
	return vp->data+ofst;
}

VFILE *vtmp()
{
	VFILE *new=(VFILE *) malloc(sizeof(VFILE));

	new->alloc=0;
	new->size=0;
	new->left=0;
	new->lv=0;
	new->vpage=0;
	new->addr=-1;
	return enqueb(VFILE, link, &vfiles, new);
}

void vclose(VFILE *vfile)
{
	VPAGE *vp, *pp;
	int x;

	if (vfile->vpage)
		vunlock (vfile->vpage);
	free (deque(VFILE, link, vfile));
	for (x=0; x!=HTSIZE; x++)
		for (pp=(VPAGE *) (htab+x), vp=pp->next; vp; )
			if (vp->vfile==vfile)
			{
				pp->next=vp->next;
				vp->next=freepages;
				freepages=vp;
				vp=pp->next;
			}
			else
				pp=vp, vp=vp->next;
	return;
}

int vallocm(VFILE *vfile, int size)
{
	int start=vsize(vfile);

	vfile->alloc=start+size;
	if (vfile->lv)
	{
		if (vheader(vfile->vpage)->addr+PGSIZE>vfile->alloc)
			vfile->lv=PGSIZE-(vfile->alloc-vheader(vfile->vpage)->addr);
		else
			vfile->lv=0;
	}
	return start;
}
